// Soldering iron controller
// David Pilling August 2017




#include <stdarg.h>


#ifdef DEBUGMODE


void pr(char *fmt, ... )
{
  char buf[128]; // resulting string limited to 128 chars

  va_list args;
  va_start (args, fmt );
  vsnprintf(buf, 128, fmt, args);
  va_end (args);

 Serial.println(buf);
}

#else

void pr(char *fmt, ... )
{
  
}

#endif

//==================================================================

unsigned int target;
unsigned int poweron;
unsigned int current;

const int powerpin=6;
const int lightpin=5;

#define MINTEMP 0
#define MAXTEMP 500


//==================================================================

int           pendingupdate=0;
unsigned long updatetime;

#define UPDATETIMEOUT 5000

typedef struct bootblockstr
{
 unsigned int target;



  
} bootblockstr;


#include <EEPROM.h>


void writebootblock(void)
{
 bootblockstr bootblock;  
 char * p=(char*)&bootblock;
 int j;

 bootblock.target=target;
 
 for(j=0;j<sizeof(bootblockstr);j++)
 {
  EEPROM.write(j,*p);
  p++;
 }
}


void readbootblock(void)
{
 bootblockstr bootblock;  
 char * p=(char*)&bootblock;
 int j;
 
 for(j=0;j<sizeof(bootblockstr);j++)
 {
  *p++=EEPROM.read(j);
 }

 target=bootblock.target;
 if(target<MINTEMP) target=MINTEMP;
 if(target>MAXTEMP) target=MAXTEMP;
}


void updatebootblock(void)
{
 pendingupdate=1;
 updatetime=millis();
}

void pollbootblock(void)
{
 if(pendingupdate)
 {
  if((millis()-updatetime)>UPDATETIMEOUT)
  {
   writebootblock();
   pendingupdate=0;
  }
 }
}

//==================================================================

/*


************************************
* This section contains coefficients for type K thermocouples for
* the two subranges of temperature listed below.  The coefficients 
* are in units of °C and mV and are listed in the order of constant 
* term up to the highest order.  The equation below 0 °C is of the form 
* E = sum(i=0 to n) c_i t^i.
*
* The equation above 0 °C is of the form 
* E = sum(i=0 to n) c_i t^i + a0 exp(a1 (t - a2)^2).
*
*     Temperature Range (°C)
*        -270.000 to 0.000 
*         0.000 to 1372.000
************************************
name: reference function on ITS-90
type: K
temperature units: °C
emf units: mV
range: -270.000, 0.000, 10
  0.000000000000E+00
  0.394501280250E-01
  0.236223735980E-04
 -0.328589067840E-06
 -0.499048287770E-08
 -0.675090591730E-10
 -0.574103274280E-12
 -0.310888728940E-14
 -0.104516093650E-16
 -0.198892668780E-19
 -0.163226974860E-22
range: 0.000, 1372.000, 9
 -0.176004136860E-01
  0.389212049750E-01
  0.185587700320E-04
 -0.994575928740E-07
  0.318409457190E-09
 -0.560728448890E-12
  0.560750590590E-15
 -0.320207200030E-18
  0.971511471520E-22
 -0.121047212750E-25
exponential:
 a0 =  0.118597600000E+00
 a1 = -0.118343200000E-03
 a2 =  0.126968600000E+03



************************************
* This section contains coefficients of approximate inverse 
* functions for type K thermocouples for the subranges of 
* temperature and voltage listed below. The range of errors of 
* the approximate inverse function for each subrange is also given. 
* The coefficients are in units of °C and mV and are listed in 
* the order of constant term up to the highest order.
* The equation is of the form t_90 = d_0 + d_1*E + d_2*E^2 + ... 
*     + d_n*E^n,
* where E is in mV and t_90 is in °C.
*
*    Temperature        Voltage            Error 
*      range              range            range
*      (°C)               (mV)             (° C)
*    -200. to 0.      -5.891 to 0.000    -0.02 to 0.04
*     0. to 500.      0.000 to 20.644    -0.05 to 0.04
*     500. to 1372.   20.644 to 54.886   -0.05 to 0.06
********************************************************
Inverse coefficients for type K:
 
Temperature  -200.             0.           500.
  Range:        0.           500.          1372.
 
  Voltage   -5.891          0.000         20.644
  Range:     0.000         20.644         54.886
 
         0.0000000E+00  0.000000E+00 -1.318058E+02
         2.5173462E+01  2.508355E+01  4.830222E+01
        -1.1662878E+00  7.860106E-02 -1.646031E+00
        -1.0833638E+00 -2.503131E-01  5.464731E-02
        -8.9773540E-01  8.315270E-02 -9.650715E-04
        -3.7342377E-01 -1.228034E-02  8.802193E-06
        -8.6632643E-02  9.804036E-04 -3.110810E-08
        -1.0450598E-02 -4.413030E-05  0.000000E+00
        -5.1920577E-04  1.057734E-06  0.000000E+00
         0.0000000E+00 -1.052755E-08  0.000000E+00
 
  Error      -0.02          -0.05          -0.05
  Range:      0.04           0.04           0.06


*/




#include <stdio.h>
#include <math.h>


float icoef[10]=
{
0.000000E+00,
2.508355E+01,
7.860106E-02,
-2.503131E-01,
8.315270E-02,
-1.228034E-02,
9.804036E-04,
-4.413030E-05,
1.057734E-06,
-1.052755E-08
};



float temperature(float mv)
{
 int i;
 float sum=0.0;
 float v=1.0;
 
 for(i=0;i<10;i++)
 {
  sum+=icoef[i]*v;
  v*=mv;
 }
 return(sum);    
}


float coef[10]=
{
 -0.176004136860E-01,
  0.389212049750E-01,
  0.185587700320E-04,
 -0.994575928740E-07,
  0.318409457190E-09,
 -0.560728448890E-12,
  0.560750590590E-15,
 -0.320207200030E-18,
  0.971511471520E-22,
 -0.121047212750E-25
};
 
 
float a0 =  0.118597600000E+00;
float a1 = -0.118343200000E-03;
float a2 =  0.126968600000E+03;


float mv(float t)
{
 int i;
 float sum=0.0;
 float v=1.0;
 
 for(i=0;i<10;i++)
 {
  sum+=coef[i]*v;
  v*=t;
 }
 
 sum+=a0*exp(a1*(t - a2)*(t-a2));
 
 return(sum);
}



//==================================================================
// Free run differential ADC read from interrupts
// save last BUFFSIZE values and average
// 

#include "DifferentialADC.c"

//http://www.avrfreaks.net/forum/any-neat-sign-extension-tricks-10-bit-16-bit

volatile long adcvalue=0;

#define ADCSHIFT 10

ISR(ADC_vect)
{
 long temp;
 long temp2;
 temp=adcvalue-(adcvalue>>ADCSHIFT);
 temp2=(long)(ADCL | (ADCH << 8));
 temp2=(temp2 ^ 512)-512;
 adcvalue=temp+temp2;
}


void analogsetup()
{
  int val;

  pr("analog setup do three reads of Vcc to stabilise");
// Note have to do this before setting off interrupts

  analogReferenceDiff(0x1); // Vcc reference

  val = analogVoltageSetup();
  pr("Vcc %d", val);
  val = analogVoltageSetup();
  pr("Vcc %d", val);
  val = analogVoltageSetup();
  pr("Vcc %d", val);

  analogSetDiffCode(analogGetCode(A4, A5, 200));
  sbi(ADCSRA, ADEN);  // enable ADC
  sbi(ADCSRA, ADATE);  // enable auto trigger
  ADCSRB &= ~(_BV(ADTS3) | _BV(ADTS2) | _BV(ADTS1) | _BV(ADTS0)); // free run mode
  ADCSRA |= (_BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0)); // divide by 128 = 9.6 KHz sample rate
  sbi(ADCSRA, ADIE); // enable ADC interrupt
  sbi(ADCSRA, ADSC); // boot
}


//==================================================================


// pin 7 switch input
#define switchinput 7

#define SWITCH_OPEN 0
#define SWITCH_CLOSED 1
#define SWITCH_BOUNCING 2

volatile int senseflag=0;
unsigned long sensetime=0;
int sensestate=SWITCH_OPEN;


void pinsense() 
{
 senseflag=1;
}

void pinsensesetup(void)
{
 pinMode(switchinput, INPUT_PULLUP);     // set pin to input
 attachInterrupt(digitalPinToInterrupt(switchinput), pinsense, CHANGE);
}

#define DEBOUNCEDELAY 50

void pinsensehandler(void)
{
 if(senseflag)
 {
  if(sensestate==SWITCH_OPEN)
  {
   poweron^=1;
   digitalWrite(lightpin,poweron?LOW:HIGH);
  }
  sensetime=millis();
  sensestate=SWITCH_BOUNCING;
  senseflag=0;
 } 

 if(sensestate==SWITCH_BOUNCING && ((millis()-sensetime)>DEBOUNCEDELAY))
 {
  sensestate=digitalRead(switchinput)?SWITCH_OPEN:SWITCH_CLOSED;
 }  
}  




//==================================================================


#include "Encoder.h"

Encoder myEnc(2, 3);

long oldPosition;

// pin 2 encoder phase 0
// pin 3 encoder phase 1
// pin 4 set to 0 v encoder earth
#define encoderzero 4


void encodersetup(void)
{
 pinMode(encoderzero, OUTPUT);           // set pin to output
 digitalWrite(encoderzero, LOW);
  

 oldPosition=myEnc.read();
}




//==================================================================

// LCD5110_TinyFont_View 
// Copyright (C)2015 Rinky-Dink Electronics, Henning Karlsen. All right reserved
// web: http://www.RinkyDinkElectronics.com/
//
// This program is a demo of the tiniest font.
//
// This program requires a Nokia 5110 LCD module.
//
// It is assumed that the LCD module is connected to
// the following pins using a levelshifter to get the
// correct voltage to the module.
//      SCK  - Pin 8
//      MOSI - Pin 9
//      DC   - Pin 10
//      RST  - Pin 11
//      CS   - Pin 12
//
#include <LCD5110_Graph.h>

LCD5110 myGLCD(8,9,10,11,12);

extern uint8_t BigNumbers[];
extern uint8_t MediumNumbers[];

void LCDsetup()
{
  myGLCD.InitLCD();
  myGLCD.setFont(BigNumbers);
}

void LCDupdate()
{
 char s[16];
 
 myGLCD.clrScr();
 myGLCD.setFont(BigNumbers);
 sprintf(s,"%d",current);
 myGLCD.print(s, CENTER, 0);

 myGLCD.setFont(MediumNumbers);
 sprintf(s,"%d",target);
 myGLCD.print(s, CENTER, 32);
 
 myGLCD.update();
}


//==================================================



void setup()
{
#ifdef DEBUGMODE  
 Serial.begin(9600);
 while(!Serial);
#endif

 target=100;
 poweron=0;
 readbootblock();
 
 analogsetup();

 pinMode(powerpin, OUTPUT);           // set pin to output
 digitalWrite(powerpin, LOW);

 pinMode(lightpin, OUTPUT);           // set pin to output
 digitalWrite(lightpin, HIGH);
 
 LCDsetup();
 encodersetup();
 pinsensesetup();
 LCDupdate();
}


void loop()
{
 int val;
 int lcdupdate=0;
 static int count=0;
 
 val=(adcvalue>>ADCSHIFT);

// pr("adc %d",val);
 
 val=analogVoltageValue(val,10,200);

// pr("volts %d",val);

 val=(int)temperature(((float)val)/10.0);

 if(val<target && poweron)  digitalWrite(powerpin, HIGH);
 else                       digitalWrite(powerpin, LOW);

 count++;
 if(count>100)
 {
  pr("%d/%d sw %d st %d",val,target,poweron,sensestate);
  count=0;
  if(val!=current)
  {
   current=val;
   lcdupdate=1; 
  }
 }

 long newPosition = myEnc.read();
 if (newPosition != oldPosition) 
 {
  long temp;
  temp=target+(newPosition-oldPosition);
  if(temp>=MINTEMP && temp<=MAXTEMP) target=temp;
  oldPosition = newPosition;
  lcdupdate=1;
  updatebootblock();
 }

 pinsensehandler(); 
 if(lcdupdate) LCDupdate();
 pollbootblock();
}


